/*
 * nwpower - NetWinder button, power and fan control
 *
 * Copyright (c) 2001 Rebel.com
 * Button driver adapted from nwbutton by Alex Holden <alex@linuxhacker.org>
 */

#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/pm.h>

#define MODULE_NAME "nwpower"
#define MODULE_VERSION "0.5"
#define DBG(x) /* x */

/* A few globals, initialized in module_init(). */

static struct pci_dev *south, *pmu;
unsigned short acpi_base;


/*
 * PART 1: BUTTON
 *
 * Check for double-click of the button and initiates shutdown
 */

#define IRQ_NETWINDER_BUTTON 13	/* Which IRQ we will register */
#define NUM_PRESSES_REBOOT 2	/* How many presses to activate shutdown */
#define BUTTON_DELAY 30		/* How many jiffies for sequence to end */

static int button_press_count;		/* The count of button presses */
static struct timer_list button_timer;	/* Times for the end of a sequence */ 

static void
button_sequence_finished (unsigned long parameters)
{
	DBG(printk ("%s: got %d button press(es)\n", 
		MODULE_NAME, button_press_count);)
	
	if (button_press_count == NUM_PRESSES_REBOOT) {
		kill_proc (1, SIGINT, 1);	/* Ask init to reboot us */
	}
	button_press_count = 0;		/* Reset the button press counter */
}

static void
button_handler (int irq, void *dev_id, struct pt_regs *regs)
{
	u8 tmp8;
	u32 tmp32;
	int button_down;

	/* Make sure interrupt was for us */
	if (!(inl (acpi_base + 0x20) & (1 << 26))) {
		printk ("%s: unknown interrupt!\n", MODULE_NAME);
		return;
	}

	/* Clear the interrupt (GPIO2) */
	outl ((1 << 26), acpi_base + 0x18);

	/* Was the button pushed or was it released? */
	pci_read_config_byte(pmu, 0x97, &tmp8);
	if (tmp8 & 0x04) {
		/* Configure GPIO2 for falling edge */
		pci_read_config_dword (pmu, 0x8c, &tmp32);
		tmp32 = (tmp32 & ~0xf) | 0x2;
		pci_write_config_dword (pmu, 0x8c, tmp32);
		DBG(printk ("%s: button pushed\n", MODULE_NAME);)
		button_down = 1;
	} else {
		/* Configure GPIO2 for rising edge */
		pci_read_config_dword (pmu, 0x8c, &tmp32);
		tmp32 = (tmp32 & ~0xf) | 0x4;
		pci_write_config_dword (pmu, 0x8c, tmp32);
		DBG(printk ("%s: button released\n", MODULE_NAME);)
		button_down = 0;
	}

	/* Only respond to button release events */
	/* In the future we might measure how long the button was down */
	/* Additional actions might be taken then. */
	if (button_down)
		return;

	/* Restart the timer */
	if (button_press_count) {
		del_timer (&button_timer);
	}
	button_press_count++;
	init_timer (&button_timer);
	button_timer.function = button_sequence_finished;
	button_timer.expires = (jiffies + BUTTON_DELAY);
	add_timer (&button_timer);
}

static int __init
init_button (void) {
	u32 tmp32;
	u16 tmp16;

	/* Register the IRQ */
	if (request_irq (IRQ_NETWINDER_BUTTON, button_handler,
			SA_INTERRUPT, "nwbutton", NULL)) {
		printk (KERN_WARNING "nwbutton: IRQ %d is not free.\n",
			IRQ_NETWINDER_BUTTON);
		return -EIO;
	}

	/* Configure GPIO2 for rising edge */
	pci_read_config_dword (pmu, 0x8c, &tmp32);
	tmp32 = (tmp32 & ~0xf) | 0x4;
	pci_write_config_dword (pmu, 0x8c, tmp32);
	DBG(pci_read_config_dword (pmu, 0x8c, &tmp32);)
	DBG(printk ("PMU_8c = %08x\n", tmp32);)

	/* Clear pending SMI/SCI for GPIO2 */
	tmp32 = (1 << 26);
	outl (tmp32, acpi_base + 0x18);
	tmp32 = inl (acpi_base + 0x18);
	DBG(printk ("ACPI+0x18 = %08x (%s)\n", tmp32,
		tmp32 & (1 << 26) ? "active" : "inactive");)

	/* Enable SMI/SCI for GPIO2 */
	tmp32 = inl (acpi_base + 0x20);
	tmp32 |= (1 << 26);
	outl (tmp32, acpi_base + 0x20);
	DBG(tmp32 = inl (acpi_base + 0x20);)
	DBG(printk ("ACPI+0x20 = %08x\n", tmp32);)

	/* Enable SCI */
	tmp16 = inw (acpi_base + 0x04);
	tmp16 |= (1 << 0);
	outw (tmp16, acpi_base + 0x04);
	DBG(tmp16 = inw (acpi_base + 0x04);)
	DBG(printk ("ACPI+0x04 = %04x\n", tmp16);)

	return 0;
}

static void __exit
cleanup_button (void) {
	u32 tmp32;
	u16 tmp16;

	/* Configure GPIO2 for no events */
	pci_read_config_dword (pmu, 0x8c, &tmp32);
	tmp32 = (tmp32 & ~0xf);
	pci_write_config_dword (pmu, 0x8c, tmp32);
	DBG(pci_read_config_dword (pmu, 0x8c, &tmp32);)
	DBG(printk ("PMU_8c = %08x\n", tmp32);)

	/* Disable SMI/SCI for GPIO2 */
	tmp32 = inl (acpi_base + 0x20);
	tmp32 &= ~(1 << 26);
	outl (tmp32, acpi_base + 0x20);
	DBG(tmp32 = inl (acpi_base + 0x20);)
	DBG(printk ("ACPI+0x20 = %08x\n", tmp32);)

	/* Clear any pending SMI/SCI for GPIO2 */
	tmp32 = (1 << 26);
	outl (tmp32, acpi_base + 0x18);
	tmp32 = inl (acpi_base + 0x18);
	DBG(printk ("ACPI+0x18 = %08x (%s)\n", tmp32,
		tmp32 & (1 << 26) ? "active" : "inactive");)

	/* Disable SCI */
	tmp16 = inw (acpi_base + 0x04);
	tmp16 &= ~(1 << 0);
	outw (tmp16, acpi_base + 0x04);
	DBG(tmp16 = inw (acpi_base + 0x04);)
	DBG(printk ("ACPI+0x04 = %04x\n", tmp16);)

	/* Unregister the IRQ */
	free_irq (IRQ_NETWINDER_BUTTON, NULL);
}


/*
 * PART 2: POWER DOWN
 *
 * At end of shutdown, turns off LEDs, parks the hard disk, etc.
 */

#define IDE0_STATUS_REG  0x1f7
#define IDE_STAT_BUSY    (1 << 7)
#define IDE_STAT_READY   (1 << 6)

#define IDE0_DEVICE_HEAD 0x1f6
#define IDE0_COMMAND_REG 0x1f7
#define IDE_CMD_SLEEP    0xe6

static void __exit
turnoff_leds (void) {
	u8 tmp8;
	struct pci_dev *dev;
	

	/* Turn off status LED (EGPIO8 and EGPIO9) */
	pci_read_config_byte (pmu, 0xc3, &tmp8);
	tmp8 &= ~(0x03);
	pci_write_config_byte (pmu, 0xc3, tmp8);

	/*
	 * Turn of ethernet LEDs by powering down the device.
	 * We ask for deep sleep state (D3).
	 */
	dev = NULL;
	while ((dev = pci_find_device (PCI_VENDOR_ID_REALTEK,
			PCI_DEVICE_ID_REALTEK_8139, dev)) != NULL) {
		DBG (printk ("Powering down device %02x.%x\n",
			PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn));)
		pci_set_power_state (dev, 3);
	}
}

static void __exit
spindown_harddisk (void) {

	/* Check for hard disk present (hackish) */
	if (inb(IDE0_STATUS_REG) == 0x7f) {
		DBG(printk("Apparently no hard disk present\n");)
		return;
	}

	/* Wait for disk to not be busy */
	DBG (printk("waiting for BSY\n");)
	while ((inb(IDE0_STATUS_REG) & IDE_STAT_BUSY) != 0);

	/* Wait for disk to indicate RDY */
	DBG (printk("waiting for RDY\n");)
	while ((inb(IDE0_STATUS_REG) & IDE_STAT_READY) == 0);

	/* Select master drive (slave would be 0xb0) */
	outb (0xa0, IDE0_DEVICE_HEAD);
	
	/* Issue the SLEEP command (E6) */
	outb (IDE_CMD_SLEEP, IDE0_COMMAND_REG);
	DBG (printk("Sleep done\n");)

	/* Wait for command to complete */
	DBG (printk("waiting for BSY\n");)
	while ((inb(IDE0_STATUS_REG) & IDE_STAT_BUSY) != 0);

	/* Read the status register */
	DBG (printk("Status reg = %02x\n", inb(IDE0_STATUS_REG));)
}

static void __exit
beep (int freq, int length) {
	unsigned int count;

	/* Stop counter */
	outb(inb(0x61) & ~0x01, 0x61);

	/* Set the frequency */
	count = 1191380 / freq;
	outb(0xb6, 0x43);
	outb(count & 0xff, 0x42);
	outb(count >> 8, 0x42);

	/* Start counter */
	outb(inb(0x61) | 0x01, 0x61);

	/* Start sound */
	outb(inb(0x61) | 0x02, 0x61);

	/* Delay */
	mdelay (length);

	/* Stop the sound */
	outb(inb(0x61) & ~0x02, 0x61);
}

static void __exit
make_shutdown_sounds (void) {
	beep (2050, 100);
	mdelay (25);
	beep (2050, 100);
}

static void __exit
shutoff_power (void) {
	/*
	 * On NetWinder Rackmounts, there is a power control board
	 * tied to GPIO22.  When this signal goes low for more than
	 * about 1/2 second, the main power is switched off.
	 */
	u8 tmp8;
	u32 tmp32;
	
	/* Make GPIO22 an output */
	pci_read_config_dword (south, 0x78, &tmp32);
	tmp32 |= (1 << 15);
	tmp32 &= ~(1 << 14);
	pci_write_config_dword (south, 0x78, tmp32);

	/* Ensure the output is low */
	pci_read_config_byte (pmu, 0xb7, &tmp8);
	tmp8 &= ~(1 << 6);
	pci_write_config_byte (pmu, 0xb7, tmp8);
}

static void __exit
turnoff_fan (void) {
	/* Turn off the fan(s) */
	u32 tmp32;

	pci_read_config_dword (pmu, 0xc8, &tmp32);
	tmp32 &= ~(1 << 24);	/* FANOUT2 */
	tmp32 &= ~(1 << 25);	/* FANOUT1 */
	pci_write_config_dword (pmu, 0xc8, tmp32);
}

static void __exit
netwinder_power_off (void) {
	DBG(printk("netwinder_power_off\n");)
	turnoff_leds();
	spindown_harddisk();
	make_shutdown_sounds();
	turnoff_fan();
	shutoff_power();
}

static void __init
init_powerdown (void) {
	pm_power_off = netwinder_power_off;
}

static void __exit
cleanup_powerdown (void) {
	pm_power_off = NULL;
}


/*
 * MODULE INIT/CLEANUP STUFF
 */

static int __init
init_nwpower (void)
{
	int ret;

	printk (KERN_INFO "%s driver version %s "
		"(C) Rebel.com <ralphs@netwinder.org>\n",
		MODULE_NAME, MODULE_VERSION);

	if (!(south = pci_find_slot(0, PCI_DEVFN(0x7, 0)))) {
		printk("error: southbridge not found\n");
		return 1;
	}

	if (!(pmu = pci_find_slot(0, PCI_DEVFN(0x11, 0)))) {
		printk("error: PMU not found\n");
		return 1;
	}

	pci_read_config_word (pmu, 0xe0, &acpi_base);
	if (acpi_base == 0) {
		printk ("error: ACPI I/O space not mapped!\n");
		return 1;
	}

	if ((ret = init_button()))
		return ret;

	init_powerdown();

	return 0;
}

static void __exit
cleanup_nwpower (void)
{
	cleanup_powerdown();
	cleanup_button();
}

#ifdef MODULE
module_init(init_nwpower);
module_exit(cleanup_nwpower);
#endif
